/*!
	\file Contour2D.cpp
	\brief Contour2D class methods definition and implementation

	Information :
		- Author	: Vincent PRAT (vprat@ifrance.com)
		- Created	: 10-07-2001 (dd-mm-yyyy)
		- Modified	: 13-07-2001 (dd-mm-yyyy)

	Copyright  2001 Vincent PRAT
*/

#include "../headers/Contour2D.h" // Contour2D

int Contour2D::FindNearestPoint( 
								float u, float v, 
								float uMin, float uMax, 
								float vMin, float vMax )
{
	// Minimal distance found (initialised to an impossible value)
	float distMin = 500000; 

	// Current distance
	float d;
	int i, res = 0;
	
	// Go through every point
	for ( i=0; i<=nTesselation; i++ )
	{
		if ( (pPoints[i].x>=uMin) && (pPoints[i].y>=vMin) &&
			 (pPoints[i].x<=uMax) && (pPoints[i].y<=vMax) )
		{
			// Compute distance between the two points
			d = sqrt( (u-pPoints[i].x)*(u-pPoints[i].x) + (v-pPoints[i].y)*(v-pPoints[i].y) );
			
			if ( d<distMin )
			{
				// We found a closer point, memorize this point and the new minimal distance
				distMin = d;
				res = i;
			}
		}
	}

	return res;
}

bool Contour2D::IsInBB( float u, float v )
{
	// Simple test for a 2D rectangle
	return (
		( u <= BBMax.x  ) && 
		( v <= BBMax.y  ) && 
		( u >= BBMin.x  ) && 
		( v >= BBMin.y  ) );
}

bool Contour2D::IsInside( float u, float v )
{
	int i, j;
	bool state = 0;
	
	// First check if we are in the bounding box
	if ( IsInBB( u, v ) )
	{
		// We are in the bounding box
		// See comp.graphics.algorithms FAQ for detail about this algorithm
		for (i = 0, j = nTesselation-1; i < nTesselation; j = i++) 
		{
			if ( 
				( ((pPoints[i].y<=v) && (v<pPoints[j].y)) ||
				  ((pPoints[j].y<=v) && (v<pPoints[i].y)) ) &&
				(u < (pPoints[j].x - pPoints[i].x)*(v - pPoints[i].y)/(pPoints[j].y - pPoints[i].y) + pPoints[i].x) )
			{
				state = !state;
			}
		}
	}

	return state; 
}

Contour2D::Contour2D( 
		int _nOrder, 
		int _nNumberOfControlPoints, Vector4d *_pControlPoints, 
		float* _pfKnots )
{
	nOrder = _nOrder;
	nNumberOfControlPoints = _nNumberOfControlPoints;

	pControlPoints = new Vector4d[ nNumberOfControlPoints ];
	memcpy( pControlPoints, _pControlPoints, nNumberOfControlPoints*sizeof(Vector4d) );

	pfKnots = new float[ nOrder + nNumberOfControlPoints ];
	memcpy( pfKnots, _pfKnots, (nOrder + nNumberOfControlPoints)*sizeof(float) );

	Initialise();
}

void Contour2D::Initialise()
{
	// Initialise the values not given at construction time
	nTesselation		= DEFAULT_TESSELATION_VALUE;
	nDegree				= nOrder - 1;
	nNumberOfKnots		= nOrder + nNumberOfControlPoints;
	nNumberOfKnotSpans	= nNumberOfKnots - 2*nDegree;

	// Allocate memory for arrays
	pfBasisCoefficients = NULL;
	pnTessKnotSpan		= NULL;
	pSurfacePoints		= NULL;
	pfBasisValues		= NULL;
	pPoints				= NULL;

	pfBasisCoefficients = new float[ nOrder*nOrder*nNumberOfKnotSpans ];
	pfBasisValues		= new float[ nOrder*(nTesselation+1) ];
	pnTessKnotSpan		= new int[ nTesselation+1 ];
	pPoints				= new Vector2d[ nTesselation+1 ];
	pSurfacePoints		= new Vector3d[ nTesselation+1 ];

	// Initialise bounding box coordinates to impossible values.
	BBMax.x = -50000;
	BBMax.y = -50000;
	BBMin.x = 50000;
	BBMin.y = 50000;

	// Pre-weight control points
	for (int i=0; i<nNumberOfControlPoints; i++)
	{
		pControlPoints[i].x *= pControlPoints[i].w;
		pControlPoints[i].y *= pControlPoints[i].w;
		pControlPoints[i].z *= pControlPoints[i].w;
	}

	// Evaluate the basis coefficients
	EvaluateBasisFunctionsCoefficients();

	// Then the basis function values at tesselated points
	EvaluateBasisFunctions();

	// Tesselate the curve
	Tesselate();
}

Contour2D::~Contour2D()
{
	// Free the memory allocated dynamically
	delete pfBasisCoefficients;
	delete pfBasisValues;
	delete pPoints;
	delete pSurfacePoints;
	delete pControlPoints;
	delete pfKnots;
	delete pnTessKnotSpan;
}

void Contour2D::EvaluateBasisFunctionsCoefficients()
{
	int i, j, k;
	int offset = 0;

	// For each knot span
	for ( i=0; i<nNumberOfKnotSpans; i++ )
	{
		// For m_nOrder polynomials
		for ( j=0; j<nOrder; j++)
		{
			// For each coefficient of the polynomial
			for ( k=0; k<nOrder; k++)
			{
				// offset calculation can be optimized.
				// I left it like that for clarity
				offset =  ( i*nOrder + j )*nOrder + k;
				pfBasisCoefficients[ offset ] = ComputeCoefficient( i+nDegree, i+j, nDegree, k );
			}
		}
	}
}

float Contour2D::ComputeCoefficient( int nInterval, int i, int p, int k )
{
	float fResult = 0.0f;

	// Very simple recursive function, exactly based on the formula
	// giving the coefficients

	if ( p==0 )
	{
		if ( i==nInterval )
		{
			fResult = 1.0f;
		}
	}
	else if ( k==0 )
	{
		if ( pfKnots[i+p]!=pfKnots[i] )
		{
			fResult -= pfKnots[i]*ComputeCoefficient( nInterval, i, p-1, 0 )/(pfKnots[i+p] - pfKnots[i]);
		}
		if ( pfKnots[i+p+1]!=pfKnots[i+1] )
		{
			fResult += pfKnots[i+p+1]*ComputeCoefficient( nInterval, i+1, p-1, 0 )/( pfKnots[i+p+1] - pfKnots[i+1] );
		}
	}
	else if ( k==p )
	{
		if ( pfKnots[i+p]!=pfKnots[i] )
		{
			fResult += ComputeCoefficient( nInterval, i, p-1, p-1 )/(pfKnots[i+p] - pfKnots[i]);
		}
		if ( pfKnots[i+p+1]!=pfKnots[i+1] )
		{
			fResult -= ComputeCoefficient( nInterval, i+1, p-1, p-1 )/( pfKnots[i+p+1] - pfKnots[i+1] );
		}
	}
	else if ( k>p )
	{
		fResult = 0.0f;
	}
	else
	{
		float C1, C2;

		if ( pfKnots[i+p]!=pfKnots[i] )
		{
			C1 = ComputeCoefficient( nInterval, i, p-1, k-1 );
			C2 = ComputeCoefficient( nInterval, i, p-1, k );

			fResult += ( C1 - pfKnots[i]*C2 )/( pfKnots[i+p]-pfKnots[i] );
		}
		if ( pfKnots[i+p+1]!=pfKnots[i+1] )
		{
			C1 = ComputeCoefficient( nInterval, i+1, p-1, k-1 );
			C2 = ComputeCoefficient( nInterval, i+1, p-1, k );

			fResult -= ( C1 - pfKnots[i+p+1]*C2 )/( pfKnots[i+p+1] - pfKnots[i+1] );
		}
	}

	return fResult;
}

void Contour2D::EvaluateBasisFunctions()
{
	int i, j, k, idx;
	float u, uInc;

	// Current knot span
	idx		= 0;

	// Initial value of u
	u		= pfKnots[nDegree];

	// u increment
	uInc	= ( pfKnots[nNumberOfKnots-nOrder] - pfKnots[nDegree])/(nTesselation);

	// For each tesselated point
	for ( i=0; i<=nTesselation; i++ )
	{
		// Find the next valid knot span
		while ( (idx<nNumberOfKnots-2*nOrder) && (u>=pfKnots[idx+nOrder]) )
		{
			idx++;
		}

		// Memorise it
		pnTessKnotSpan[i] = idx+nDegree;

		// Evaluate the basis function for this value of u
		// We use Horner's method to evaluated a polynomial
		for ( j=0; j<nOrder; j++)
		{
			pfBasisValues[ i*nOrder+j ] = pfBasisCoefficients[ (idx*nOrder+j)*nOrder + nDegree ];

			for ( k=nDegree-1; k>=0; k-- )
			{
				pfBasisValues[ i*nOrder+j ] = 
					pfBasisValues[ i*nOrder+j ] * u + 
					pfBasisCoefficients[ (idx*nOrder+j)*nOrder + k ];
			}
		}

		// Next paramter value
		u += uInc;
	}
}

void Contour2D::Tesselate()
{
	int u, i;
	int nCurrentKnotSpan;
	int nBasisOffset, nControlPointOffset;
	float dBasisValue;
	float rhw; 
	Vector4d oTemp;
	
	// For each point of the mesh
	for ( u=0; u<=nTesselation; u++ )
	{
		// Get current knot span
		nCurrentKnotSpan = pnTessKnotSpan[u];
		
		// offset in the precalculated basis functions array and control points array
		nBasisOffset = u*nOrder;
		nControlPointOffset = nCurrentKnotSpan - nDegree;

		// For each control point having influence,
		// Get basis value and blend control points
		// Remember, control points are pre-weighted

		dBasisValue = pfBasisValues[ nBasisOffset ];
		
		oTemp.x = dBasisValue*pControlPoints[nControlPointOffset].x;
		oTemp.y = dBasisValue*pControlPoints[nControlPointOffset].y;
		oTemp.w = dBasisValue*pControlPoints[nControlPointOffset].w;

		for ( i=0; i<nOrder; i++ )
		{
			dBasisValue = pfBasisValues[ nBasisOffset+i ];

			oTemp.x += dBasisValue*pControlPoints[nControlPointOffset+i].x;
			oTemp.y += dBasisValue*pControlPoints[nControlPointOffset+i].y;
			oTemp.w += dBasisValue*pControlPoints[nControlPointOffset+i].w;
		}
		rhw = 1.0f/oTemp.w;

		// Save tesselated point
		pPoints[u].x = rhw * oTemp.x;
		pPoints[u].y = rhw * oTemp.y;

		// Build bounding rectangle
		if (pPoints[u].x < BBMin.x)
			BBMin.x = pPoints[u].x;
		if (pPoints[u].y < BBMin.y)
			BBMin.y = pPoints[u].y;
		if (pPoints[u].x > BBMax.x)
			BBMax.x = pPoints[u].x;
		if (pPoints[u].y > BBMax.y)
			BBMax.y = pPoints[u].y;
	}
}

void Contour2D::ComputeSurfacePoints( TrimmedNURBSSurface *surface )
{
	int i, k, l;
	Vector4d oTemp;
	int nControlPointOffset;
	float fUBasisValue, fVBasisValue, rhw;

	// For each contour point
	for ( i=0; i<=nTesselation; i++ )
	{
		nControlPointOffset = 0;

		memset( &oTemp, 0, sizeof(Vector4d) );
		
		// For each control point in the V direction
		for (k=0; k<surface->anNumberOfControlPoints[1]; k++)
		{
			// Compute basis value for the parameter value pPoint[i].y
			fVBasisValue = surface->EvaluateBasisFunction( k, surface->anDegree[1], pPoints[i].y, surface->pfVKnots );
			
			// For each control point in the U direction
			for (l=0; l<surface->anNumberOfControlPoints[0]; l++)
			{
				// Evaluate basis function value for the parameter value pPoint[i].x
				fUBasisValue = surface->EvaluateBasisFunction( l, surface->anDegree[0], pPoints[i].x, surface->pfUKnots );
				
				// do the sum for coordinates (control points are pre-weighted)
				oTemp.x += fVBasisValue*fUBasisValue*surface->pControlPoints[ nControlPointOffset ].x;
				oTemp.y += fVBasisValue*fUBasisValue*surface->pControlPoints[ nControlPointOffset ].y;
				oTemp.z += fVBasisValue*fUBasisValue*surface->pControlPoints[ nControlPointOffset ].z;
				oTemp.w += fVBasisValue*fUBasisValue*surface->pControlPoints[ nControlPointOffset ].w;
				
				// Next control point
				nControlPointOffset++;
			}
		}
		rhw = 1.0f/oTemp.w;
		
		// Save coordinates of computed point
		pSurfacePoints[i].x = rhw * oTemp.x;
		pSurfacePoints[i].y = rhw * oTemp.y;
		pSurfacePoints[i].z = rhw * oTemp.z;
	}
}






























